"""
============================================
vidgear library code is placed under the MIT license
Copyright (c) 2019 Abhishek Thakur

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
===============================================
"""

# Contains all the support functions/modules required by Vidgear 

# import the neccesary packages
import os, sys
import cv2
from pkg_resources import parse_version



def check_python_version():
	"""
	returns current python version's - first bit 
	"""
	return sys.version_info[0]



def check_CV_version():
	"""
	returns current OpenCV binary version's first bit 
	"""
	if parse_version(cv2.__version__) >= parse_version('4'):
		return 4
	else:
		return 3



def capPropId(property):
	"""
	Retrieves the Property's Integer(Actual) value. 
	"""
	return getattr(cv2, property)



def dict2Args(param_dict):
	"""
	converts dict to list(args)
	"""
	args = []
	for key in param_dict.keys():
		args.append(key)
		args.append(param_dict[key])
	return args



def get_valid_ffmpeg_path(custom_ffmpeg = '', is_windows = False, ffmpeg_download_path = '', logging = False):
	"""
	Validate the FFmpeg path/binaries and returns valid FFmpeg file executable location(also downloads static binaries on windows) 
	"""
	final_path = ''
	if is_windows:
		#checks if current os is windows
		if custom_ffmpeg:
			#if custom FFmpeg path is given assign to local variable
			final_path += custom_ffmpeg
		else:
			#otherwise auto-download them
			try:
				if ffmpeg_download_path:
					#checks if FFmpeg download path specified
					pass
				else:
					#otherwise save to Temp Directory
					import tempfile
					ffmpeg_download_path = tempfile.gettempdir()

				if logging:
					print('FFmpeg Windows Download Path: {}'.format(ffmpeg_download_path))

				#download Binaries
				_path = download_ffmpeg_binaries(path = ffmpeg_download_path, os_windows = is_windows)
				#assign to local variable
				final_path += _path

			except Exception as e:
				#log if any error occured
				if logging:
					print(e)
					print('Error downloading FFmpeg binaries, Check your network and Try again!')
				return False
		if os.path.isfile(final_path):
			#check if valid FFmpeg file exist
			pass
		elif os.path.isfile(os.path.join(final_path, 'ffmpeg.exe')):
			#check if FFmpeg directory exists, if does, then check for valid file 
			final_path = os.path.join(final_path, 'ffmpeg.exe')
		else:
			#else return False
			if logging:
				print('No valid FFmpeg executables found at Custom FFmpeg path!')
			return False
	else:
		#otherwise perform test for Unix
		if custom_ffmpeg:
			#if custom FFmpeg path is given assign to local variable
			if os.path.isfile(custom_ffmpeg):
				#check if valid FFmpeg file exist
				final_path += custom_ffmpeg
			elif os.path.isfile(os.path.join(custom_ffmpeg, 'ffmpeg')):
				#check if FFmpeg directory exists, if does, then check for valid file 
				final_path = os.path.join(custom_ffmpeg, 'ffmpeg')
			else:
				#else return False
				if logging:
					print('No valid FFmpeg executables found at Custom FFmpeg path!')
				return False
		else:
			#otherwise assign ffmpeg binaries from system
			final_path += "ffmpeg"

	if logging:
		print('Final FFmpeg Path: {}'.format(final_path))

	# Final Auto-Validation for FFmeg Binaries. returns final path if test is passed
	if validate_ffmpeg(final_path, logging = logging):
		return final_path
	else:
		return False



def download_ffmpeg_binaries(path, os_windows = False):
	"""
	Download and Extract FFmpeg Static Binaries for windows(if not available)
	"""
	import platform
	final_path = ''
	if os_windows:
		windows_bit = 'win64' if platform.machine().endswith('64') else 'win32' #checks current Windows Bit Mode
		#inialize varibles
		file_url = 'https://ffmpeg.zeranoe.com/builds/{}/static/ffmpeg-latest-{}-static.zip'.format(windows_bit, windows_bit)
		file_name = os.path.join(os.path.abspath(path),'ffmpeg-latest-{}-static.zip'.format(windows_bit))
		file_path = os.path.join(os.path.abspath(path), 'ffmpeg-latest-{}-static/bin/ffmpeg.exe'.format(windows_bit))
		base_path, _ = os.path.split(file_name) #extract file base path
		#check if file already exists
		if os.path.isfile(file_path):
			final_path += file_path #skip download if does 
		else:
			import requests
			import zipfile
			#check if given pth has write access
			assert os.access(path, os.W_OK), "Permission Denied: Cannot write ffmpeg binaries to directory = " + path
			#remove leftovers
			if os.path.isfile(file_name):
				os.remove(file_name)
			#download and write file to the given path
			with open(file_name, "wb") as f:
				print("No Custom FFmpeg path provided, Auto-Downloading binaries for Windows. Please wait...")
				response  = requests.get(file_url, stream=True)
				total_length = response.headers.get('content-length')
				if total_length is None: # no content length header
					f.write(response.content)
				else:
					dl = 0
					total_length = int(total_length)
					for data in response.iter_content(chunk_size=4096):
						dl += len(data)
						f.write(data)
						done = int(50 * dl / total_length)
						sys.stdout.write("\r[{}{}]{}{}".format('=' * done, ' ' * (50-done), done * 2, '%') )    
						sys.stdout.flush()
			print("\nExtracting executables, Please Wait...")
			with zipfile.ZipFile(file_name, "r") as zip_ref:
				zip_ref.extractall(base_path)
			#perform cleaning
			os.remove(file_name)
			print("FFmpeg binaries for Windows Configured Successfully!")
			final_path += file_path
	#return final path
	return final_path



def validate_ffmpeg(path, logging = False):
	"""
	Validate FFmeg Binaries. returns True if tests passed
	"""
	try:
		#get the FFmpeg version
		version = check_output([path, "-version"])
		firstline = version.split(b'\n')[0]
		version = firstline.split(b' ')[2].strip()
		if logging:
			#log if test are passed 
			print('FFmpeg validity Test Passed!')
			print('Found valid FFmpeg Version: `{}` installed on this system'.format(version))
	except Exception as e:
		#log if test are failed
		if logging:
			print(e)
			print('FFmpeg validity Test Failed!')
		return False
	return True



def check_output(*args, **kwargs):
	"""
	return output from the sub-process
	"""
	#silent subprocess execution
	closeNULL = 0
	import subprocess as sp
	try:
		from subprocess import DEVNULL
		closeNULL = 0
	except ImportError:
		DEVNULL = open(os.devnull, 'wb')
		closeNULL = 1
	#execute command in subprocess
	process = sp.Popen(stdout=sp.PIPE, stderr=DEVNULL, *args, **kwargs)
	output, unused_err = process.communicate()
	retcode = process.poll()
	#close the process
	if closeNULL:
		DEVNULL.close()
	#if error occured raise error
	if retcode:
		cmd = kwargs.get("args")
		if cmd is None:
			cmd = args[0]
		error = sp.CalledProcessError(retcode, cmd)
		error.output = output
		raise error
	return output


# This is a modified algorithm/code based on findings of `Simple video stabilization using OpenCV` 
# published on February 20, 2014 by nghiaho12 (http://nghiaho.com/?p=2093)

"""
============================================
vidgear library code is placed under the MIT license
Copyright (c) 2019 Abhishek Thakur

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
===============================================
"""

# import the necessary packages
from collections import deque
from .helper import check_CV_version
import numpy as np
import cv2



class Stabilizer:
	
	"""
	This is an auxiliary class that enables Real-time Video Stabilization for vidgear with minimalistic latency and at the expense of little to no 
	additional computational power requirement. The basic idea behind it is to tracks and save the salient feature array for the given number of frames 
	and then uses these anchor point to cancel out all perturbations relative to it for the incoming frames in the queue. This class relies heavily on 
	Threaded Queue mode for error-free, ultrafast frame handling and it is enabled by default.

	:param smoothing_radius` (int) : to alter averaging window size. It handles the quality of stabilization at expense of latency and sudden panning. 
									/ Larger its value, less will be panning, more will be latency and vice-versa. It's default value is 25.
	:param border_size (int) : to alter output border cropping. It's will crops the border to reduce the black borders from stabilization being too noticeable. 
								/ Larger its value, more will be cropping. It's default value is 0 (i.e. no cropping).			
	:param border_type (string) : to change the border mode. Valid border types are 'black', 'reflect', 'reflect_101', 'replicate' and 'wrap'. It's default value is 'black'

	:param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.

	"""
	
	def __init__(self, smoothing_radius = 25, border_type = 'black', border_size = 0 , logging = False):

		# initialize deques for handling input frames and its indexes
		self.frame_queue = deque(maxlen=smoothing_radius)
		self.frame_queue_indexes = deque(maxlen=smoothing_radius)

		# define and create Adaptive histogram equalization (AHE) object for optimizations
		self.clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8,8))

		# initialize global vars
		self.smoothing_radius = smoothing_radius #averaging window, handles the quality of stabilization at expense of latency and sudden panning
		self.smoothed_path = None #handles the smoothed path with box filter
		self.path = None #handles path i.e cumulative sum of pevious_2_current transformations along a axis
		self.transforms = [] #handles pevious_2_current transformations [dx,dy,da]
		self.frame_transforms_smoothed = None #handles smoothed array of pevious_2_current transformations w.r.t to frames
		self.previous_gray = None #handles previous gray frame
		self.previous_keypoints = None #handles previous detect_GFTTed keypoints w.r.t previous gray frame
		self.frame_height, self.frame_width = (0, 0) #handles width and height of input frames

		# define valid border modes
		border_modes = {'black': cv2.BORDER_CONSTANT,'reflect': cv2.BORDER_REFLECT, 'reflect_101': cv2.BORDER_REFLECT_101, 'replicate': cv2.BORDER_REPLICATE, 'wrap': cv2.BORDER_WRAP}
		# choose valid border_mode from border_type
		if border_type in ['black', 'reflect', 'reflect_101', 'replicate', 'wrap']:
			#initialize global border mode variable 
			self.border_mode = border_modes[border_type]
		else:
			#otherwise log if not
			if logging:
				print('Invalid input border type!')
			self.border_mode = border_modes['black'] #reset to default mode
			
		# define normalized box filter
		self.box_filter = np.ones(smoothing_radius)/smoothing_radius

		# define cropping factor, Crops the border to reduce the black borders from stabilization being too noticeable.
		self.border_crop_size = border_size

		#decide whether to log
		self.logging = logging



	def stabilize(self, frame):
		"""
		Return stabilized video frame

		:param frame(numpy.ndarray): input video frame
		"""
		#check if frame is None
		if frame is None:
			#return if it does
			return
		
		# initiate transformations capturing
		if not self.frame_queue:
			#for first frame
			previous_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY) #convert to gray
			previous_gray = self.clahe.apply(previous_gray) #optimize gray frame
			self.previous_keypoints = cv2.goodFeaturesToTrack(previous_gray, maxCorners=200, qualityLevel=0.05, minDistance=30.0, blockSize=3 , mask=None, useHarrisDetector=False, k=0.04) #track features using GFTT
			self.frame_height, self.frame_width = frame.shape[:2] #save input frame height and width
			self.frame_queue.append(frame) #save frame to deque
			self.frame_queue_indexes.append(0) #save frame index to deque
			self.previous_gray = previous_gray[:] #save gray frame clone for further processing

		elif self.frame_queue_indexes[-1] <= self.smoothing_radius - 1:
			#for rest of frames 
			self.frame_queue.append(frame) #save frame to deque
			self.frame_queue_indexes.append(self.frame_queue_indexes[-1]+1) #save frame index
			self.generate_transformations() #generate transformations
			if self.frame_queue_indexes[-1] == self.smoothing_radius - 1:
				#calculate smooth path once transformation capturing is completed
				for i in range(3):
					#apply normalized box filter to the path
				   self.smoothed_path[:,i] = self.box_filter_convolve((self.path[:,i]), window_size=self.smoothing_radius)
				#calculate deviation of path from smoothed path
				deviation = self.smoothed_path - self.path
				#save smoothed transformation
				self.frame_transforms_smoothed = self.frame_transform + deviation
		else:
			#start applying transformations
			self.frame_queue.append(frame) #save frame to deque
			self.frame_queue_indexes.append(self.frame_queue_indexes[-1]+1) #save frame index
			self.generate_transformations() #generate transformations
			#calculate smooth path once transformation capturing is completed
			for i in range(3):
				#apply normalized box filter to the path
			   self.smoothed_path[:,i] = self.box_filter_convolve((self.path[:,i]), window_size=self.smoothing_radius)
			#calculate deviation of path from smoothed path
			deviation = self.smoothed_path - self.path
			#save smoothed transformation
			self.frame_transforms_smoothed = self.frame_transform + deviation
			#return transformation applied stabilized frame
			return self.apply_transformations()



	def generate_transformations(self):
		"""
		Generate pevious_2_current transformations [dx,dy,da]
		"""
		frame_gray = cv2.cvtColor(self.frame_queue[-1], cv2.COLOR_BGR2GRAY) #retrieve current frame and convert to gray
		frame_gray = self.clahe.apply(frame_gray) #optimize it

		#calculate optical flow using Lucas-Kanade differential method
		curr_kps, status, error = cv2.calcOpticalFlowPyrLK(self.previous_gray, frame_gray, self.previous_keypoints, None)
		
		#select only valid keypoints
		valid_curr_kps = curr_kps[status==1] #current
		valid_previous_keypoints = self.previous_keypoints[status==1] #previous
		
		#calculate optimal affine transformation between pevious_2_current keypoints
		if check_CV_version() == 3:
			#backward compatibility with OpenCV3 
			transformation = cv2.estimateRigidTransform(valid_previous_keypoints,valid_curr_kps, False)
		else:
			transformation = cv2.estimateAffinePartial2D(valid_previous_keypoints,valid_curr_kps)[0]
		
		#check if transformation is not None
		if not(transformation is None):
			# pevious_2_current translation in x direction
			dx = transformation[0, 2]
			# pevious_2_current translation in y direction
			dy = transformation[1, 2]
			# pevious_2_current rotation in angle
			da = np.arctan2(transformation[1, 0], transformation[0, 0])
		else:
			#otherwise zero it
			dx = dy = da = 0

		#save this transformation
		self.transforms.append([dx, dy, da])

		#calculate path from cumulative transformations sum
		self.frame_transform = np.array(self.transforms, dtype='float32')
		self.path = np.cumsum(self.frame_transform, axis=0)
		#create smoothed path from a copy of path
		self.smoothed_path=np.copy(self.path)

		#re-calculate and save GFTT keypoints for current gray frame
		self.previous_keypoints = cv2.goodFeaturesToTrack(frame_gray, maxCorners=200, qualityLevel=0.05, minDistance=30.0, blockSize=3 , mask=None, useHarrisDetector=False, k=0.04)
		#save this gray frame for further processing
		self.previous_gray = frame_gray[:]



	def box_filter_convolve(self, path, window_size):
		"""
		applies normalized linear box filter to path w.r.t to averaging window
		
		:param: path(nd.array)(cumulative sum of transformations)
		:param: averaging window size(int)
		"""
		#pad path to size of averaging window
		path_padded = np.pad(path, (window_size, window_size), 'median')
		#apply linear box filter to path
		path_smoothed = np.convolve(path_padded, self.box_filter, mode='same')
		#crop the smoothed path to original path 
		path_smoothed = path_smoothed[window_size:-window_size]
		#assert if cropping is completed
		assert path.shape == path_smoothed.shape
		#return smoothed path
		return path_smoothed



	def apply_transformations(self):
		"""
		Applies affine transformation to the given frame 
		build from previously calculated transformations
		"""

		#extract frame and its index from deque
		queue_frame = self.frame_queue.popleft()
		queue_frame_index = self.frame_queue_indexes.popleft()

		#create border around extracted frame w.r.t border_size
		bordered_frame = cv2.copyMakeBorder(queue_frame, top=self.border_crop_size, bottom=self.border_crop_size, left=self.border_crop_size, right=self.border_crop_size, borderType=self.border_mode, value=[0, 0, 0])
		alpha_bordered_frame = cv2.cvtColor(bordered_frame, cv2.COLOR_BGR2BGRA) #create alpha channel
		#extract alpha channel
		alpha_bordered_frame[:, :, 3] = 0 
		alpha_bordered_frame[self.border_crop_size:self.border_crop_size + self.frame_height, self.border_crop_size:self.border_crop_size + self.frame_width, 3] = 255

		#extracting Transformations w.r.t frame index
		dx = self.frame_transforms_smoothed[queue_frame_index,0] #x-axis
		dy = self.frame_transforms_smoothed[queue_frame_index,1] #y-axis
		da = self.frame_transforms_smoothed[queue_frame_index,2] #angle

		#building 2x3 transformation matrix from extracted transformations
		queue_frame_transform = np.zeros((2,3), np.float32)
		queue_frame_transform[0,0] = np.cos(da)
		queue_frame_transform[0,1] = -np.sin(da)
		queue_frame_transform[1,0] = np.sin(da)
		queue_frame_transform[1,1] = np.cos(da)
		queue_frame_transform[0,2] = dx
		queue_frame_transform[1,2] = dy

		#Applying an affine transformation to the frame
		frame_wrapped = cv2.warpAffine(alpha_bordered_frame, queue_frame_transform, alpha_bordered_frame.shape[:2][::-1], borderMode=self.border_mode)

		#drop alpha channel
		frame_stabilized = frame_wrapped[:, :, :3]

		#finally return stabilized frame
		return frame_stabilized


	def clean(self):
		"""
		clean resources(deque)
		"""
		# check if deque present
		if self.frame_queue:
			#clear frame deque
			self.frame_queue.clear()
			#clear frame indexes deque
			self.frame_queue_indexes.clear()





	


"""
============================================
vidgear library code is placed under the MIT license
Copyright (c) 2019 Abhishek Thakur

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
===============================================
"""

# import the necessary packages
from .camgear import CamGear



class VideoGear:

	"""
	VideoGear API provides a special internal wrapper around VidGear's exclusive Video Stabilizer class. Furthermore, VideoGear API can provide 
	internal access to both CamGear and PiGear APIs separated by a special flag. Thereby, this API holds the exclusive power for any incoming 
	VideoStream from any source, whether it is live or not, to stabilize it directly with minimum latency and memory requirements.

	VideoGear Specific parameters:
	
		:param (boolean) enablePiCamera: set this flag to access PiGear or CamGear class respectively. 
										/ This means the if enablePiCamera flag is `True`, PiGear class will be accessed 
										/ and if `False`, the camGear Class will be accessed. Its default value is False.

		:param (boolean) stabilize: set this flag to enable access to VidGear's Stabilizer Class. This basically enables(if True) or disables(if False) 
										video stabilization in VidGear. Its default value is False.

		:param (dict) **options: can be used in addition, to pass parameter supported by VidGear's stabilizer class.
								/ Supported dict keys are: 
									- `SMOOTHING_RADIUS` (int) : to alter averaging window size. It handles the quality of stabilization at expense of latency and sudden panning. 
															/ Larger its value, less will be panning, more will be latency and vice-versa. It's default value is 25.
									- `BORDER_SIZE` (int) : to alter output border cropping. It's will crops the border to reduce the black borders from stabilization being too noticeable. 
															/ Larger its value, more will be cropping. It's default value is 0 (i.e. no cropping).			
									- `BORDER_TYPE` (string) : to change the border mode. Valid border types are 'black', 'reflect', 'reflect_101', 'replicate' and 'wrap'. It's default value is 'black'
		
		:param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.
	
	CamGear Specific supported parameters for VideoGear:

		:param source : take the source value for CamGear Class. Its default value is 0. Valid Inputs are:
			- Index(integer): Valid index of the video device.
			- YouTube Url(string): Youtube URL as input.
			- Network_Stream_Address(string): Incoming Stream Valid Network address. 
			- GStreamer (string) videostream Support
		:param (boolean) y_tube: enables YouTube Mode in CamGear Class, i.e If enabled the class will interpret the given source string as YouTube URL. 
								/ Its default value is False.
		:param (int) backend: set the backend of the video stream (if specified). Its default value is 0.


	PiGear Specific supported parameters for VideoGear:
	
		:param (tuple) resolution: sets the resolution (width,height) in Picamera class. Its default value is (640,480).
		:param (integer) framerate: sets the framerate in Picamera class. Its default value is 25.


	Common parameters for CamGear and PiGear: 
		:param (dict) **options: sets parameter supported by PiCamera or Camgear (whichever being accessed) Class to the input video stream. 
								/ These attribute provides the flexibity to manuplate input raspicam video stream directly. 
								/ Parameters can be passed using this **option, allows you to pass keyworded variable length of arguments to given Class.
		:param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.
		:param (integer) time_delay: sets time delay(in seconds) before start reading the frames. 
							/ This delay is essentially required for camera to warm-up. 
							/ Its default value is 0.
	"""

	def __init__(self, enablePiCamera = False, stabilize = False, source=0, y_tube = False, backend = 0, colorspace = None, resolution=(640, 480), framerate=25, logging = False, time_delay = 0, **options):
		
		self.stablization_mode = stabilize

		if self.stablization_mode:
			from .stabilizer import Stabilizer
			s_radius, border_size, border_type = (25, 0, 'black')
			if options:
				if "SMOOTHING_RADIUS" in options:
					if isinstance(options["SMOOTHING_RADIUS"],int):
						s_radius = options["SMOOTHING_RADIUS"] #assigsn special parameter to global variable
					del options["SMOOTHING_RADIUS"] #clean
				if "BORDER_SIZE" in options:
					if isinstance(options["BORDER_SIZE"],int):
						border_size = options["BORDER_SIZE"] #assigsn special parameter
					del options["BORDER_SIZE"] #clean
				if "BORDER_TYPE" in options:
					if isinstance(options["BORDER_TYPE"],str):
						border_type = options["BORDER_TYPE"] #assigsn special parameter
					del options["BORDER_TYPE"] #clean
			self.stabilizer_obj = Stabilizer(smoothing_radius = s_radius, border_type = border_type, border_size = border_size, logging = logging)
			#log info
			if logging:
				print('Enabling Stablization Mode for the current video source!')

		if enablePiCamera:
			# only import the pigear module only if required
			from .pigear import PiGear

			# initialize the picamera stream by enabling PiGear Class
			self.stream = PiGear(resolution=resolution, framerate=framerate, colorspace = colorspace, logging = logging, time_delay = time_delay, **options)

		else:
			# otherwise, we are using OpenCV so initialize the webcam
			# stream by activating CamGear Class
			self.stream = CamGear(source=source, y_tube = y_tube, backend = backend, colorspace = colorspace, logging = logging, time_delay = time_delay, **options)

	def start(self):
		# start the threaded video stream
		self.stream.start()
		return self

	def update(self):
		# grab the next frame from the stream
		self.stream.update()

	def read(self):
		# return the current frame
		while self.stablization_mode:
			frame = self.stream.read()
			if frame is None:
				break
			frame_stab = self.stabilizer_obj.stabilize(frame)
			if not(frame_stab is None):
				return frame_stab
		return self.stream.read()

	def stop(self):
		# stop the thread and release any resources
		self.stream.stop()
		if self.stablization_mode:
			self.stabilizer_obj.clean()


"""
============================================
vidgear library code is placed under the MIT license
Copyright (c) 2019 Abhishek Thakur

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
===============================================
"""

# import the packages
from threading import Thread
from pkg_resources import parse_version
import traceback
from .helper import capPropId



try:
	# import OpenCV Binaries
	import cv2

	# check whether OpenCV Binaries are 3.x+
	if parse_version(cv2.__version__) >= parse_version('3'):
		pass
	else:
		raise ImportError('OpenCV library version >= 3.0 is only supported by this library')

except ImportError as error:
	raise ImportError('Failed to detect OpenCV executables, install it with `pip install opencv-contrib-python` command.')



class PiGear:

	"""
	PiGear is similar to CamGear but exclusively made to support various Raspberry Pi Camera Modules 
	(such as OmniVision OV5647 Camera Module and Sony IMX219 Camera Module). To interface with these 
	modules correctly, PiGear provides a flexible multi-threaded wrapper around complete picamera 
	python library and provides us the ability to exploit its various features like brightness, saturation, sensor_mode, etc. effortlessly.
	
	:param (tuple) resolution: sets the resolution (width,height). Its default value is (640,480).

	:param (integer) framerate: sets the framerate. Its default value is 25.

	:param (string) colorspace: set colorspace of the video stream. Its default value is None.

	:param (dict) **options: sets parameter supported by PiCamera Class to the input video stream. 
							/ These attribute provides the flexibility to manipulate input raspicam video stream directly. 
							/ Parameters can be passed using this **option, allows you to pass key worded variable length of arguments to PiGear Class.

	:param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.

	:param (integer) time_delay: sets time delay(in seconds) before start reading the frames. 
					/ This delay is essentially required for camera to warm-up. 
					/ Its default value is 0.

	"""
	
	def __init__(self, resolution=(640, 480), framerate=25, colorspace = None, logging = False, time_delay = 0, **options):

		try:
			import picamera
			from picamera.array import PiRGBArray
			from picamera import PiCamera
			#print(cv2.__version__)
		except ImportError as error:
			# Output expected ImportErrors.
			raise ImportError('Failed to detect Picamera executables, install it with "pip install picamera" command.')

		# initialize the picamera stream
		self.camera = PiCamera()
		self.camera.resolution = resolution
		self.camera.framerate = framerate

		#initialize framerate variable
		self.framerate = framerate

		#initializing colorspace variable
		self.color_space = None

		#reformat dict
		options = {k.strip(): v for k,v in options.items()}

		try: 
			# apply attributes to source if specified
			for key, value in options.items():
				setattr(self.camera, key, value)

			# separately handle colorspace value to int conversion
			if not(colorspace is None):
				self.color_space = capPropId(colorspace.strip())

		except Exception as e:
			# Catch if any error occurred
			if logging:
				print(e)

		# enable rgb capture array thread and capture stream
		self.rawCapture = PiRGBArray(self.camera, size=resolution)
		self.stream = self.camera.capture_continuous(self.rawCapture,format="bgr", use_video_port=True)

		#frame variable initialization		
		for stream in self.stream:
			self.frame = stream.array
			self.rawCapture.seek(0)
			self.rawCapture.truncate()
			break

		# applying time delay to warm-up picamera only if specified
		if time_delay:
			import time
			time.sleep(time_delay)

		#thread initialization
		self.thread = None

		# enable logging if specified
		self.logging = logging

		# initialize termination flag
		self.terminate = False



	def start(self):
		"""
		start the thread to read frames from the video stream
		"""
		self.thread = Thread(target=self.update, args=())
		self.thread.daemon = True
		self.thread.start()
		return self



	def update(self):
		"""
		Update frames from stream
		"""
		# keep looping infinitely until the thread is terminated
		try:
			for stream in self.stream:
			# grab the frame from the stream and clear the stream in
			# preparation for the next frame
				if stream is None:
					if self.logging:
						print('The Camera Module is not working Properly!')
					self.terminate = True
				if self.terminate:
					break
				frame = stream.array
				self.rawCapture.seek(0)
				self.rawCapture.truncate()

				if not(self.color_space is None):
					# apply colorspace to frames
					color_frame = None
					try:
						if isinstance(self.color_space, int):
							color_frame = cv2.cvtColor(frame, self.color_space)
						else:
							self.color_space = None
							if self.logging:
								print('Colorspace value {} is not a valid Colorspace!'.format(self.color_space))
								
					except Exception as e:
						# Catch if any error occurred
						self.color_space = None
						if self.logging:
							print(e)
							print('Input Colorspace is not a valid Colorspace!')

					if not(color_frame is None):
						self.frame = color_frame
					else:
						self.frame = frame
				else:
					self.frame = frame

		except Exception as e:
			if self.logging:
				print(traceback.format_exc())
			self.terminate =True
			pass

		# release picamera resources
		self.stream.close()
		self.rawCapture.close()
		self.camera.close()



	def read(self):
		"""
		return the frame
		"""
		return self.frame



	def stop(self):
		"""
		Terminates the Read process
		"""
		# indicate that the thread should be terminated
		self.terminate = True
		# wait until stream resources are released (producer thread might be still grabbing frame)
		if self.thread is not None: 
			self.thread.join()
			#properly handle thread exit




"""
============================================
vidgear library code is placed under the MIT license
Copyright (c) 2019 Abhishek Thakur

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
===============================================
"""

# import the necessary packages/libraries
from pkg_resources import parse_version
import os, sys, time
import subprocess as sp

from .helper import get_valid_ffmpeg_path
from .helper import capPropId
from .helper import dict2Args



try:
	# import OpenCV Binaries
	import cv2
	# check whether OpenCV Binaries are 3.x+
	if parse_version(cv2.__version__) >= parse_version('3'):
		pass
	else:
		raise ImportError('OpenCV library version >= 3.0 is only supported by this library')
except ImportError as error:
	raise ImportError('Failed to detect OpenCV executables, install it with `pip install opencv-contrib-python` command.')



class WriteGear:

	"""
	WriteGear solely handles various powerful FFmpeg tools that allow us to do almost anything you can imagine with multimedia files. 
	With WriteGear API, you can process real-time video frames into a lossless format and specification suitable for our playback in 
	just a few lines of codes. These specifications include setting bitrate, codec, framerate, resolution, subtitles, compression, etc. 
	Furthermore, we can multiplex extracted audio at the output with compression and all that in real-time(see this example). 
	In addition to this, WriteGear also provides flexible access to OpenCV's VideoWriter API which provide some basic tools 
	for video frames encoding but without compression.

	WriteGear primarily operates in the following two modes:

		Compression Mode: In this mode, WriteGear utilizes FFmpeg's inbuilt encoders to encode lossless multimedia files. 
							It provides us the ability to exploit almost any available parameters available within FFmpeg, with so much ease 
							and flexibility and while doing that it robustly handles all errors/warnings quietly. You can find more about this mode here.

		Non-Compression Mode: In this mode, WriteGear utilizes basic OpenCV's inbuilt VideoWriter API. Similar to compression mode, WriteGear also supports 
							all parameters manipulation available within OpenCV's VideoWriter API. But this mode lacks the ability to manipulate encoding parameters 
							and other important features like video compression, audio encoding, etc. You can learn about this mode here.


	```Warning: In case, This class fails to detect valid FFmpeg executables on your system, It can automatically fallbacks to Non-Compression Mode.```


	:param output_filename (string): sets the output Video file path/filename. Valid Inputs are:

		- path: Valid path of the directory to save the output video file. Then, the class will automatically assign filename(with default extension `.mp4`).  

		- filename(with path): Valid filename(with valid extension) of the output video file. 

	```Note: Make sure to provide valid filename with path and file-extension(based on encoder(in case of FFmpeg)/fourcc(in case of OpenCV) being currently used)```

	:param compression_mode (boolean): enables/disables Video Compression Mode, i.e If enabled(means Compression Mode) WriteGear class will utilize 
						/ the FFmpeg installed/given executables to encode output video and if disabled(means Non-Compression Mode) 
						/ OpenCV's VideoWriter Class will be used. Its default value is True.

	:param custom_ffmpeg (string): (if specified) sets the custom path/directory where the `Custom FFmpeg executables` are located, works only in Compression Mode.
						/ It is compulsory to provide custom binaries on a Windows Machine, otherwise this class with automatically download and extract suitable Static FFmpeg 
						/ binaries to Temporary directory.Read VidGear Docs for more info.

	:param logging (boolean): set this flag to enable/disable error logging essential for debugging. Its default value is False.

	:param **output_params (dict): sets all properties supported by FFmpeg(in Compression Mode)/ OpenCV's VideoWriter(in Non-Compression Mode) properties 
						/ to the input video stream in CamGear Class. All the valid User-defined FFmeg/OpenCV's VideoWriter output video file parameters/properties 
						/ manipulation is currently supported. These attribute provides the flexibility to manipulate output video  directly. 
						/ In addition other special attributes to manipulate WriteGear inbuilt properties are also supported. Read VidGear Docs for more info.

	"""
	
	def __init__(self, output_filename = '', compression_mode = True , custom_ffmpeg = '', logging = False, **output_params):

		# assign parameter values to class variables
		self.compression = compression_mode
		self.os_windows  = True if os.name == 'nt' else False #checks if machine in-use is running windows or not
		self.logging = logging

		# initialize various important class variables
		self.output_parameters = {}
		self.inputheight = None
		self.inputwidth = None
		self.inputchannels = None
		self.inputframerate = 0
		self.output_dimensions = None
		self.process = None #handle process to be frames written
		self.DEVNULL = None #handles silent execution of FFmpeg (if logging is disabled)
		self.cmd = ''     #handle FFmpeg Pipe command
		self.ffmpeg = ''  #handle valid FFmpeg binaries location
		self.initiate = True #initiate one time process for valid process initialization


		# handles output file name (if not given)
		if not output_filename:
			raise ValueError('Kindly provide a valid `output_filename` value, Refer VidGear Docs for more information!')
		elif output_filename and os.path.isdir(output_filename): # check if directory path is given instead
			output_filename = os.path.join(output_filename, 'VidGear-{}.mp4'.format(time.strftime("%Y%m%d-%H%M%S"))) # auto-assign valid name and adds it to path
		else:
			pass

		# some definitions and assigning output file absolute path to class variable 
		_filename = os.path.abspath(output_filename)
		self.out_file = _filename
		basepath, _ = os.path.split(_filename) #extract file base path for debugging ahead


		if output_params:
			#handle user defined output dimensions(must be a tuple or list)
			if output_params and "-output_dimensions" in output_params:
				self.output_dimensions = output_params["-output_dimensions"] #assign special parameter to global variable
				del output_params["-output_dimensions"] #clean
			#cleans and reformat output parameters
			try:
				self.output_parameters = {str(k).strip().lower(): str(v).strip() for k,v in output_params.items()}
			except Exception as e:
				if self.logging:
					print(e)
				raise ValueError('Wrong output_params parameters passed to WriteGear class!')

		#handles FFmpeg binaries validity tests 
		if self.compression:

			if self.logging:
				print('Compression Mode is enabled therefore checking for valid FFmpeg executables!')
				print(self.output_parameters)

			# handles where to save the downloaded FFmpeg Static Binaries on Windows(if specified)
			ffmpeg_download_path_ = ''
			if self.output_parameters and "-ffmpeg_download_path" in self.output_parameters:
				ffmpeg_download_path_ += self.output_parameters["-ffmpeg_download_path"]
				del self.output_parameters["-ffmpeg_download_path"] #clean

			#handle input framerate if specified
			if self.output_parameters and "-input_framerate" in self.output_parameters:
				self.inputframerate += float(self.output_parameters["-input_framerate"])
				del self.output_parameters["-input_framerate"] #clean

			#validate the FFmpeg path/binaries and returns valid FFmpeg file executable location(also downloads static binaries on windows) 
			actual_command = get_valid_ffmpeg_path(custom_ffmpeg, self.os_windows, ffmpeg_download_path = ffmpeg_download_path_, logging = self.logging)

			#check if valid path returned
			if actual_command:
				self.ffmpeg += actual_command #assign it to class variable
				if self.logging:
					print('Found valid FFmpeg executables: `{}`'.format(self.ffmpeg))
			else:
				#otherwise disable Compression Mode
				if self.logging and not self.os_windows:
					print('Kindly install working FFmpeg or provide a valid custom FFmpeg Path')
				print('Caution: Disabling Video Compression Mode since no valid FFmpeg executables found on this machine!')
				self.compression = False # compression mode disabled

		#validate this class has the access rights to specified directory or not
		assert os.access(basepath, os.W_OK), "Permission Denied: Cannot write to directory = " + basepath

		#display confirmation if logging is enabled/disabled
		if self.compression and self.ffmpeg:
			self.DEVNULL = open(os.devnull, 'wb') 
			if self.logging:
				print('Compression Mode is configured properly!')
		else:
			if self.logging:
				print('Compression Mode is disabled, Activating OpenCV In-built Writer!')




	def write(self, frame, rgb_mode = False):

		"""  
		pipelines ndarray frames to valid Video File. Utilizes FFmpeg in Compression Mode and OpenCV VideoWriter Class in Non-Compression Mode

		:param frame (ndarray): valid frame
		:param rgb_mode (boolean): set this flag to enable rgb_mode, i.e. specifies that incoming frames are of RGB format(instead of default BGR). Its default value is False.

		"""
		if frame is None: #NoneType frames will be skipped 
			return

		#get height, width and number of channels of current frame
		height, width = frame.shape[:2]
		channels = frame.shape[-1] if frame.ndim == 3 else 1

		# assign values to class variables on first run 
		if self.initiate:
			self.inputheight = height
			self.inputwidth = width
			self.inputchannels = channels
			if self.logging:
				print('InputFrame => Height:{} Width:{} Channels:{}'.format(self.inputheight, self.inputwidth, self.inputchannels))

		#validate size of frame
		if height != self.inputheight or width != self.inputwidth:
			raise ValueError('All frames in a video should have same size')
		#validate number of channels
		if channels != self.inputchannels:
			raise ValueError('All frames in a video should have same number of channels')

		if self.compression:
			# checks if compression mode is enabled

			#initiate FFmpeg process on first run
			if self.initiate:
				#start pre-processing and initiate process 
				self.Preprocess(channels, rgb = rgb_mode)
				# Check status of the process
				assert self.process is not None

			#write the frame
			try:
				self.process.stdin.write(frame.tostring())
			except (OSError, IOError):
				# log something is wrong!
				print ('BrokenPipeError caught: Wrong Values passed to FFmpeg Pipe, Kindly Refer Docs!')
				self.DEVNULL.close()
				raise ValueError #for testing purpose only
		else:
			# otherwise initiate OpenCV's VideoWriter Class
			if self.initiate:
				#start VideoWriter Class process
				self.startCV_Process()
				# Check status of the process
				assert self.process is not None
				if self.logging:
					# log OpenCV warning
					print('Warning: RGBA and 16-bit grayscale video frames are not supported by OpenCV yet, switch to `compression_mode` to use them!')
			#write the frame
			self.process.write(frame)



	def Preprocess(self, channels, rgb = False):
		"""
		pre-processing FFmpeg parameters
		
		:param channels (int): Number of channels
		:param rgb_mode (boolean): set this flag to enable rgb_mode, Its default value is False.
		"""

		#turn off initiate flag
		self.initiate = False
		#initialize input parameters
		input_parameters = {}

		#handle dimensions
		dimensions = ''
		if self.output_dimensions is None: #check if dimensions are given
			dimensions += '{}x{}'.format(self.inputwidth, self.inputheight) #auto derive from frame
		else:
			dimensions += '{}x{}'.format(self.output_dimensions[0],self.output_dimensions[1]) #apply if defined
		input_parameters["-s"] = str(dimensions)

		#handles pix_fmt based on channels(HACK)
		if channels == 1:
			input_parameters["-pix_fmt"] = "gray"
		elif channels == 2:
			input_parameters["-pix_fmt"] = "ya8"
		elif channels == 3:
			input_parameters["-pix_fmt"] = "rgb24" if rgb else "bgr24"
		elif channels == 4:
			input_parameters["-pix_fmt"] = "rgba" if rgb else "bgra"
		else:
			raise ValueError("Handling Frames with (1 > Channels > 4) is not implemented!")

		if self.inputframerate > 5:
			#set input framerate - minimum threshold is 5.0
			if self.logging:
				print("Setting Input FrameRate = {}".format(self.inputframerate))
			input_parameters["-framerate"] = str(self.inputframerate)

		#initiate FFmpeg process
		self.startFFmpeg_Process(input_params = input_parameters, output_params = self.output_parameters)



	def startFFmpeg_Process(self, input_params, output_params):

		"""
		Start FFmpeg process

		:param input_params (dict): Input parameters
		:param output_params (dict): Output parameters
		"""

		#convert input parameters to list
		input_parameters = dict2Args(input_params)

		#pre-assign default encoder parameters (if not assigned by user).
		if "-vcodec" not in output_params:
			output_params["-vcodec"] = "libx264"
		if output_params["-vcodec"] in ["libx264", "libx265"]:
			if "-crf" in output_params: 
				pass
			else:
				output_params["-crf"] = "18"
			if "-preset" in output_params: 
				pass
			else:
				output_params["-preset"] = "fast"
		#convert output parameters to list
		output_parameters = dict2Args(output_params)
		#format command
		cmd = [self.ffmpeg , "-y"] + ["-f", "rawvideo", "-vcodec", "rawvideo"] + input_parameters + ["-i", "-"] + output_parameters + [self.out_file]
		#assign value to class variable
		self.cmd += " ".join(cmd)
		# Launch the FFmpeg process
		if self.logging:
			print(self.cmd)
			# In debugging mode
			self.process = sp.Popen(cmd, stdin=sp.PIPE, stdout=sp.PIPE, stderr=None)
		else:
			# In silent mode
			self.process = sp.Popen(cmd, stdin=sp.PIPE, stdout=self.DEVNULL, stderr=sp.STDOUT)


	def startCV_Process(self):
		"""
		Start OpenCV VideoWriter Class process
		"""

		#turn off initiate flag
		self.initiate = False

		#initialize essential parameter variables
		FPS = 0
		BACKEND = ''
		FOURCC = 0
		COLOR = True



		#pre-assign default encoder parameters (if not assigned by user).
		if "-fourcc" not in self.output_parameters:
			FOURCC = cv2.VideoWriter_fourcc(*"MJPG")
		if "-fps" not in self.output_parameters:
			FPS = 25

		#auto assign dimensions	
		HEIGHT = self.inputheight
		WIDTH = self.inputwidth

		#assign parameter dict values to variables
		try:
			for key, value in self.output_parameters.items():
				if key == '-fourcc':
					FOURCC = cv2.VideoWriter_fourcc(*(value.upper()))
				elif key == '-fps':
					FPS = float(value)
				elif key =='-backend' and value.upper() in ['CAP_FFMPEG','CAP_GSTREAMER']:
					BACKEND = capPropId(value.upper())
				elif key == '-color':
					COLOR = bool(int(value))
				else:
					pass

		except Exception as e:
			# log if something is wrong
			if self.logging:
				print(e)
			raise ValueError('Wrong Values passed to OpenCV Writer, Kindly Refer Docs!')

		if self.logging:
			#log values for debugging
			print('FILE_PATH: {}, FOURCC = {}, FPS = {}, WIDTH = {}, HEIGHT = {}, BACKEND = {}'.format(self.out_file,FOURCC, FPS, WIDTH, HEIGHT, BACKEND))

		#start different process for with/without Backend.
		if BACKEND: 
			self.process = cv2.VideoWriter(self.out_file, apiPreference = BACKEND, fourcc = FOURCC, fps = FPS, frameSize = (WIDTH, HEIGHT), isColor = COLOR)
		else:
			self.process = cv2.VideoWriter(self.out_file, fourcc = FOURCC, fps = FPS, frameSize = (WIDTH, HEIGHT), isColor = COLOR)


	def close(self):
		"""
		Terminates the Write process
		"""
		if self.compression:
			#if Compression Mode is enabled
			if self.process is None:  
				return  #no process was initiated at first place
			if self.process.poll() is not None: 
				return  # process was already dead
			if self.process.stdin:
				self.process.stdin.close() #close `stdin` output
			if self.output_parameters and "-i" in self.output_parameters:
				self.process.terminate()
				self.process.wait() #wait if still process is still processing some information
				self.process = None 
				self.DEVNULL.close() #close it
			else:
				self.process.wait() #wait if still process is still processing some information
				self.process = None 
				self.DEVNULL.close() #close it
		else:
			#if Compression Mode is disabled
			if self.process is None: 
				return  #no process was initiated at first place
			self.process.release() #close it


	






		



"""
============================================
vidgear library code is placed under the MIT license
Copyright (c) 2019 Abhishek Thakur

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
===============================================
"""

# import the necessary packages
from threading import Thread
from pkg_resources import parse_version
import numpy as np
import time



try:
	# import OpenCV Binaries
	import cv2
	# check whether OpenCV Binaries are 3.x+
	if parse_version(cv2.__version__) >= parse_version('3'):
		pass
	else:
		raise ImportError('OpenCV library version >= 3.0 is only supported by this library')
except ImportError as error:
	raise ImportError('Failed to detect OpenCV executables, install it with `pip install opencv-contrib-python` command.')



class NetGear:

	"""
	NetGear is exclusively designed to transfer video frames synchronously between interconnecting systems over the network in real-time. 
	This is achieved by implementing a high-level wrapper around PyZmQ python library that contains python bindings for ZeroMQ - a 
	high-performance asynchronous distributed messaging library that aim to be used in distributed or concurrent applications. 
	It provides a message queue, but unlike message-oriented middleware, a ZeroMQ system can run without a dedicated message broker. 
	Furthermore, NetGear currently supports two ZeroMQ messaging patterns: i.e zmq.PAIR and zmq.REQ and zmq.REP and the supported 
	protocol are: 'tcp', 'upd', 'pgm', 'inproc', 'ipc'.

	Threaded Queue Mode => Sequentially adds and releases frames to/from deque and handles overflow of this queue. It utilizes 
	deques that support thread-safe, memory efficient appends and pops from either side of the deque with approximately the 
	same O(1) performance in either direction.  

	:param address(string): sets the valid network address of the server/client. Network addresses unique identifiers across the 
							network. Its default value of this parameter is based on mode it is working, 'localhost' for Send Mode
							and `*` for Receive Mode.

	:param port(string): sets the valid network port of the server/client. A network port is a number that identifies one side 
							of a connection between two devices on network. It is used determine to which process or application 
							a message should be delivered. Its default value is `5555` .

	:param protocol(string): sets the valid messaging protocol between server and client. A network protocol is a set of established rules
							 that dictates how to format, transmit and receive data so computer network devices - from servers and 
							 routers to endpoints - can communicate regardless of the differences in their underlying infrastructures, 
							 designs or standards. Supported protocol are: 'tcp', 'upd', 'pgm', 'inproc', 'ipc'. Its default value is `tcp`.

	:param pattern(int): sets the supported messaging pattern(flow of communication) between server and client. Messaging patterns are network 
							oriented architectural pattern that describes the flow of communication between interconnecting systems.
							vidgear provides access to ZeroMQ's pre-optimized sockets which enables you to take advantage of these patterns.
							The supported patterns are:
								0: zmq.PAIR -> In this, the communication is bidirectional. There is no specific state stored within the socket. 
												There can only be one connected peer.The server listens on a certain port and a client connects to it.
								1. zmq.REQ/zmq.REP -> In this, ZMQ REQ sockets can connect to many servers. The requests will be
														interleaved or distributed to both the servers. socket zmq.REQ will block 
														on send unless it has successfully received a reply back and socket zmq.REP 
														will block on recv unless it has received a request.
							Its default value is `0`(i.e zmq.PAIR).

	:param (boolean) receive_mode: set this flag to select the Netgear's Mode of operation. This basically activates `Receive Mode`(if True) and `Send Mode`(if False). 
									Furthermore `recv()` function will only works when this flag is enabled(i.e. `Receive Mode`) and `send()` function will only works 
									when this flag is disabled(i.e.`Send Mode`). Checkout VidGear docs for usage details.
									Its default value is False(i.e. Send Mode is activated by default).

	:param **options(dict): can be used to pass parameters to NetGear Class. 
							/This attribute provides the flexibility to manipulate ZeroMQ input parameters 
							/directly. Checkout vidgear docs for usage details.

	:param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.

	"""
	
	def __init__(self, address = None, port = None, protocol = None,  pattern = 0, receive_mode = False, logging = False, **options):

		try:
			#import PyZMQ library
			import zmq
			#import ZMQError
			from zmq.error import ZMQError
			#assign values to global variable
			self.zmq = zmq
			self.ZMQError = ZMQError
		except ImportError as error:
			#raise error
			raise ImportError('pyzmq python library not installed. Kindly install it with `pip install pyzmq` command.')

		#log and enable threaded queue mode
		if logging:
			print('Threaded Mode is enabled by default for NetGear.')
		#import deque
		from collections import deque
		#define deque and assign it to global var
		self.queue = deque(maxlen=96) #max len 96 to check overflow

		#define valid messaging pattern `0`: zmq.PAIR and `1`:(zmq.REQ,zmq.REP)
		valid_messaging_patterns = {0:(zmq.PAIR,zmq.PAIR), 1:(zmq.REQ,zmq.REP)}

		# initialize messaging pattern
		msg_pattern = None
		# check whether user-defined messaging pattern is valid
		if isinstance(pattern, int) and pattern in valid_messaging_patterns:
			#assign value 
			msg_pattern = valid_messaging_patterns[pattern]
			self.pattern = pattern #add it to global variable for further use
		else:
			#otherwise default to 0:`zmq.PAIR`
			self.pattern = 0
			msg_pattern = valid_messaging_patterns[self.pattern]
			if logging:
				#log it
				print('Wrong pattern, Defaulting to `zmq.PAIR`! Kindly refer Docs for more Information.')
		
		#check  whether user-defined messaging protocol is valid
		if not (protocol is None) and protocol in ['tcp', 'upd', 'pgm', 'inproc', 'ipc']:
			pass
		else:
			# else default to `tcp` protocol
			protocol = 'tcp'
			if logging:
				#log it
				print('Protocol is not valid or provided. Defaulting to `tcp` protocol! Kindly refer Docs for more Information.')


		self.msg_flag = 0 #handles connection flags
		self.msg_copy = False #handles whether to copy data
		self.msg_track = False #handles whether to track packets
		try: 
			#reformat dict
			options = {k.strip(): v for k,v in options.items()}
			# apply attributes to source if specified and valid
			for key, value in options.items():
				if key == 'flag' and isinstance(value, str):
					self.msg_flag = getattr(zmq, value)
				elif key == 'copy' and isinstance(value, bool):
					self.msg_copy = value
				elif key == 'track' and isinstance(value, bool):
					self.msg_track = value
				else:
					pass
		except Exception as e:
			# Catch if any error occurred
			if logging:
				print(e)


		# enable logging if specified
		self.logging = logging

		# initialize termination flag
		self.terminate = False

		# initialize exit_loop flag
		self.exit_loop = False

		# initialize and assign receive mode to global variable
		self.receive_mode = receive_mode

		# define messaging context
		self.msg_context = zmq.Context()

		#check whether `receive_mode` is enabled by user
		if receive_mode:
			# if does than define connection address and port
			if address is None: #define address
				address = '*'
			if port is None: #define port
				port = '5555'

			if logging:
				import random
				#generate and log random device id for server-side debugging
				self.id = ''.join(random.choice('0123456789ABCDEF') for i in range(5))
				print('This device ID is {}.'.format(self.id))

			try:
				# initialize and define thread-safe messaging socket
				self.msg_socket = self.msg_context.socket(msg_pattern[1])
				# bind socket to given protocol, address and port
				self.msg_socket.bind(protocol+'://' + str(address) + ':' + str(port))
				# define socket receive timeout
				self.msg_socket.setsockopt(zmq.LINGER, 0)
				if logging:
					#log it on success
					print('Successfully Binded to address: {}.'.format(protocol+'://' + str(address) + ':' + str(port)))
			except Exception as e:
				# otherwise raise value error
				raise ValueError('Failed to bind address: {} and pattern: {}! Kindly recheck all parameters.'.format((protocol+'://' + str(address) + ':' + str(port)), pattern))
			
			# initialize and start threading instance
			self.thread = Thread(target=self.update, args=())
			self.thread.daemon = True
			self.thread.start()
			if logging:
				#log it 
				print('Multi-threaded Receive Mode is enabled Successfully!')
		else:
			#otherwise default to `Send Mode

			# if does than define connection address and port`
			if address is None: #define address
				address = 'localhost'
			if port is None: #define port
				port = '5555'
			try:
				# initialize and define thread-safe messaging socket
				self.msg_socket = self.msg_context.socket(msg_pattern[0])
				if self.pattern == 1:
					# if pattern is 1, define additional flags
					self.msg_socket.REQ_RELAXED = True
					self.msg_socket.REQ_CORRELATE = True
				# bind socket to given protocol, address and port
				self.msg_socket.connect(protocol+'://' + str(address) + ':' + str(port))
				# define socket receive timeout
				self.msg_socket.setsockopt(zmq.LINGER, 0)
				if logging:
					#log it
					print('Successfully connected to address: {}.'.format(protocol+'://' + str(address) + ':' + str(port)))
			except Exception as e:
				# otherwise raise value error
				raise ValueError('Failed to connect address: {} and pattern: {}! Kindly recheck all parameters.'.format((protocol+'://' + str(address) + ':' + str(port)), pattern))

			if logging:
				#log it
				print('Send Mode is successfully activated and ready to send data!')



	def update(self):
		"""
		Updates recovered frames from messaging network to the queue
		"""
		# initialize frame variable
		frame = None
		# keep looping infinitely until the thread is terminated
		while not self.exit_loop:
			# check if global termination_flag is enabled  
			if self.terminate:
				# check whether there is still frames in queue before breaking out
				if len(self.queue)>0:
					continue
				else:
					break

			#check queue buffer for overflow
			if len(self.queue) < 96:
				pass
			else:
				#stop iterating if overflowing occurs
				time.sleep(0.000001)
				continue

			# extract json data out of socket
			msg_json = self.msg_socket.recv_json(flags=self.msg_flag)
			# check if terminate_flag` is enabled in json
			if msg_json['terminate_flag']:
				if self.pattern == 1:
					# if pattern is 1, then send back server the info about termination
					self.msg_socket.send_string('Termination received on port: {} !'.format(self.id))
				#assign values to global termination flag
				self.terminate = msg_json['terminate_flag']
				continue
			# extract array from socket
			msg_data = self.msg_socket.recv(flags=self.msg_flag, copy=self.msg_copy, track=self.msg_track)
			# send confirmation message to server for debugging
			self.msg_socket.send_string('Data received on port: {} !'.format(self.id))
			# recover frame from array buffer
			frame_buffer = np.frombuffer(msg_data, dtype=msg_json['dtype'])
			# reshape frame
			frame = frame_buffer.reshape(msg_json['shape'])
			#extract if any message from server and display it
			if msg_json['message']:
				print(msg_json['message'])

			# append recovered frame to queue
			self.queue.append(frame)

		# finally properly close the socket
		self.msg_socket.close()



	def recv(self):
		"""
		return the recovered frame
		"""
		# check whether `receive mode` is activated
		if self.receive_mode:
			pass
		else:
			#otherwise raise value error and exit
			raise ValueError('recv() function cannot be used while receive_mode is disabled. Kindly refer vidgear docs!')
			self.terminate = True
		# check whether or not termination flag is enabled
		while not self.terminate:
			# check if queue is empty
			if len(self.queue)>0:
				return self.queue.popleft()
			else:
				continue
		# otherwise return NoneType
		return None



	def send(self, frame, message = None):
		"""
		send the frames over the messaging network

		:param frame(ndarray): frame array to send
		:param message(string/int): additional message for the client(s)  
		"""
		# check whether `receive_mode` is disabled
		if not self.receive_mode:
			pass
		else:
			#otherwise raise value error and exit
			raise ValueError('send() function cannot be used while receive_mode is enabled. Kindly refer vidgear docs!')
			self.terminate = True

		# define exit_flag and assign value
		exit_flag = True if (frame is None or self.terminate) else False
		#check whether exit_flag is False
		if not exit_flag:
			#check whether the incoming frame is contiguous
			if frame.flags['C_CONTIGUOUS']:
				pass
			else:
				#otherwise make it contiguous
				frame = np.ascontiguousarray(frame, dtype=frame.dtype)
		# prepare the json dict and assign values
		msg_dict = dict(terminate_flag = exit_flag,
						message = message if not(message is None) else '',
						dtype = str(frame.dtype),
						shape = frame.shape)
		# send the json dict
		self.msg_socket.send_json(msg_dict, self.msg_flag|self.zmq.SNDMORE)
		# send the frame array with correct flags
		self.msg_socket.send(frame, flags = self.msg_flag, copy=self.msg_copy, track=self.msg_track)
		# wait for confirmation
		if self.logging:
			# log confirmation 
			print(self.msg_socket.recv())
		else:
			# otherwise be quiet
			self.msg_socket.recv()



	def close(self):
		"""
		Terminates the NetGear processes safely
		"""
		if self.logging:
			#log it
			print('\n Terminating various {} Processes \n'.format('Receive Mode' if self.receive_mode else 'Send Mode'))
		#  whether `receive_mode` is enabled or not
		if self.receive_mode:
			# indicate that process should be terminated
			self.terminate = True
			# check whether queue mode is empty
			if not(self.queue is None):
				self.queue.clear()
			# call immediate termination
			self.exit_loop = True
			# wait until stream resources are released (producer thread might be still grabbing frame)
			if self.thread is not None: 
				self.thread.join()
				#properly handle thread exit
		else:
			# otherwise indicate that the thread should be terminated
			self.terminate = True
			# send termination flag to client
			term_dict = dict(terminate_flag = True)
			self.msg_socket.send_json(term_dict)
			# properly close the socket
			self.msg_socket.close()


"""
============================================
vidgear library code is placed under the MIT license
Copyright (c) 2019 Abhishek Thakur

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
===============================================
"""

# import the necessary packages
from threading import Thread
from pkg_resources import parse_version
from .helper import capPropId
import numpy as np
import time



try:
	# import OpenCV Binaries
	import cv2
	# check whether OpenCV Binaries are 3.x+
	if parse_version(cv2.__version__) >= parse_version('3'):
		pass
	else:
		raise ImportError('OpenCV library version >= 3.0 is only supported by this library')

except ImportError as error:
	raise ImportError('Failed to detect OpenCV executables, install it with `pip install opencv-contrib-python` command.')



class ScreenGear:

	"""
	With ScreenGear, we can easily define an area on the computer screen or an open window to record the live screen frames in 
	real-time at the expense of insignificant latency. To achieve this, ScreenGear provides a high-level multi-threaded wrapper 
	around mss python library API and also supports the flexible direct parameter manipulation. Furthermore, ScreenGear relies on 
	Threaded Queue mode for ultra-fast live frame handling and which is enabled by default.

	Threaded Queue Mode => Sequentially adds and releases frames to/from deque and handles overflow of this queue. It utilizes 
	Deques that support thread-safe, memory efficient appends and pops from either side of the deque with approximately the 
	same O(1) performance in either direction.  


	:param monitor(int): sets the Positions/Location of monitor where to grab frame from. More information can be found in the docs. 
						/ It default value is 1 (means current monitor will be used).

	:param **options(dict): can be used to pass parameters to ScreenGear Class. 
							/This attribute provides the flexibility to manipulate mss input parameters 
							/directly like the dimensions of the region of the given monitor from where the 
							frames to be grabbed. Checkout VidGear docs for usage details.

	:param (string) colorspace: set the colorspace of the video stream. Its default value is None.

	:param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.

	"""
	
	def __init__(self, monitor = 1, colorspace = None, logging = False, **options):
		#intialize threaded queue mode
		self.threaded_queue_mode = True
		try:
			# try import necessary system specific mss library
			import platform			
			if platform.system() == 'Linux':
				from mss.linux import MSS as mss
			elif platform.system() == 'Windows':
				from mss.windows import MSS as mss
			elif platform.system() == 'Darwin':
				from mss.darwin import MSS as mss
			else:
				from mss import mss
			#import mss error handler
			from mss.exception import ScreenShotError
		except ImportError as error:
			# otherwise raise import error
			raise ImportError('python-mss library not found, install it with `pip install mss` command.')
		# create mss object
		self.mss_object = mss() 
		# create monitor instance for the user-defined monitor
		monitor_instance = None
		if (monitor > 0):
			monitor_instance = self.mss_object.monitors[monitor]
		else:
			raise ValueError("`monitor` value cannot be negative, Read Docs!")
		# Initiate User-Defined Threaded Queue Mode
		if options:
			if "THREADED_QUEUE_MODE" in options:
				if isinstance(options["THREADED_QUEUE_MODE"],bool):
					self.threaded_queue_mode = options["THREADED_QUEUE_MODE"] #assigsn special parameter to global variable
				del options["THREADED_QUEUE_MODE"] #clean
				#reformat option dict
		self.queue = None
		#intialize deque
		if self.threaded_queue_mode:
			#import deque
			from collections import deque
			#define deque and assign it to global var
			self.queue = deque(maxlen=96) #max len 96 to check overflow
			#log it
			if logging:
				print('Enabling Threaded Queue Mode!') 
		else:
			#otherwise disable it
			self.threaded_queue_mode = False
		#intiate screen dimension handler
		screen_dims = {}
		#initializing colorspace variable
		self.color_space = None
		try: 
			#reformat proper mss dict and assign to screen dimension handler
			screen_dims = {k.strip(): v for k,v in options.items() if k.strip() in ["top", "left", "width", "height"]}
			# separately handle colorspace value to int conversion
			if not(colorspace is None):
				self.color_space = capPropId(colorspace.strip())
		except Exception as e:
			# Catch if any error occurred
			if logging:
				print(e)
		# intialize mss capture instance
		self.mss_capture_instance = None
		try:
			# check whether user-defined dimensions are provided
			if screen_dims and len(screen_dims) == 4:
				self.mss_capture_instance = screen_dims #create instance from dimensions
			else:
				self.mss_capture_instance = monitor_instance #otherwise create instance from monitor
			# extract global frame from instance
			self.frame = np.asanyarray(self.mss_object.grab(self.mss_capture_instance))
			if self.threaded_queue_mode:
				#intitialize and append to queue
				self.queue.append(self.frame)
		except ScreenShotError:
			#otherwise catch and log errors
			raise ValueError("ScreenShotError caught: Wrong dimensions passed to python-mss, Kindly Refer Docs!")
			if logging:
				print(self.mss_object.get_error_details())
		# enable logging if specified
		self.logging = logging
		# thread initialization
		self.thread=None
		# initialize termination flag
		self.terminate = False


	def start(self):
		"""
		start the thread to read frames from the video stream
		"""
		self.thread = Thread(target=self.update, args=())
		self.thread.daemon = True
		self.thread.start()
		return self



	def update(self):
		"""
		Update frames from stream
		"""
		#intialize frame variable
		frame = None
		# keep looping infinitely until the thread is terminated
		while True:
			# if the thread terminate is set, stop the thread
			if self.terminate:
				break
			if self.threaded_queue_mode:
				#check queue buffer for overflow
				if len(self.queue) < 96:
					pass
				else:
					#stop iterating if overflowing occurs
					time.sleep(0.000001)
					continue
			try:
				frame = np.asanyarray(self.mss_object.grab(self.mss_capture_instance))
			except ScreenShotError:
				raise RuntimeError(self.mss_object.get_error_details())
				self.terminate = True
				continue
			if frame is None or frame.size == 0:
				#no frames received, then safely exit
				if self.threaded_queue_mode:
					if len(self.queue)>0:
						pass
					else:
						self.terminate = True
				else:
					self.terminate = True
			if not(self.color_space is None):
				# apply colorspace to frames
				color_frame = None
				try:
					if isinstance(self.color_space, int):
						color_frame = cv2.cvtColor(frame, self.color_space)
					else:
						self.color_space = None
						if self.logging:
							print('Colorspace value {} is not a valid Colorspace!'.format(self.color_space))
				except Exception as e:
					# Catch if any error occurred
					self.color_space = None
					if self.logging:
						print(e)
						print('Input Colorspace is not a valid Colorspace!')
				if not(color_frame is None):
					self.frame = color_frame
				else:
					self.frame = frame
			else:
				self.frame = frame
			#append to queue
			if self.threaded_queue_mode:
				self.queue.append(self.frame)
		# finally release mss resources
		self.mss_object.close()



	def read(self):
		"""
		return the frame
		"""
		while self.threaded_queue_mode:
			if len(self.queue)>0:
				return self.queue.popleft()
		return self.frame



	def stop(self):
		"""
		Terminates the Read process
		"""
		#terminate Threaded queue mode seperately
		if self.threaded_queue_mode and not(self.queue is None):
			self.queue.clear()
			self.threaded_queue_mode = False
			self.frame = None
		# indicate that the thread should be terminated
		self.terminate = True
		# wait until stream resources are released (producer thread might be still grabbing frame)
		if self.thread is not None: 
			self.thread.join()
			#properly handle thread exit



	





"""
============================================
vidgear library code is placed under the MIT license
Copyright (c) 2019 Abhishek Thakur

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
===============================================
"""

# import the necessary packages
from threading import Thread
from pkg_resources import parse_version
from .helper import capPropId
from .helper import check_CV_version
import re, time



#Note: Remember, Not all parameters are supported by all cameras which is 
#one of the most troublesome part of the OpenCV library. Each camera type, 
#from android cameras to USB cameras to professional
#ones offer a different interface to modify its parameters. 
#Therefore there are many branches in OpenCV code to support 
#as many of them, but of course not all possible devices 
#are covered and thereby works.

#To check parameter values supported by your webcam, you can hook your camera
#to your Linux machine and use command `v4l2-ctl -d 0 --list-formats-ext` (where 0 is index of given camera)
#to list the supported video parameters and their values.



try:
	# import OpenCV Binaries
	import cv2
	# check whether OpenCV Binaries are 3.x+
	if parse_version(cv2.__version__) >= parse_version('3'):
		pass
	else:
		raise ImportError('OpenCV library version >= 3.0 is only supported by this library')

except ImportError as error:
	raise ImportError('Failed to detect OpenCV executables, install it with `pip install opencv-contrib-python` command.')



def youtube_url_validation(url):
	"""
	convert youtube video url and checks its validity
	"""
	youtube_regex = (
		r'(https?://)?(www\.)?'
		'(youtube|youtu|youtube-nocookie)\.(com|be)/'
		'(watch\?v=|embed/|v/|.+\?v=)?([^&=%\?]{11})')
	youtube_regex_match = re.match(youtube_regex, url)
	if youtube_regex_match:
		return youtube_regex_match.group(6)
	return youtube_regex_match



class CamGear:

	"""
	CamGear API supports a diverse range of video streams which can handle/control video stream almost any IP/USB Cameras, multimedia 
	video file format (upto 4k tested), network stream URL such as http(s), rtp, rstp, mms, etc. In addition to this, it also supports 
	live Gstreamer's RAW pipelines and YouTube video/livestreams URLs. CamGear provides a flexible, high-level multi-threaded wrapper 
	around OpenCV's VideoCapture API with access almost all of its available parameters and also employs pafy's APIs for YouTube streaming. 
	Furthermore, CamGear relies exclusively on Threaded Queue mode for ultra-fast, error-free and synchronized frame handling.  

	Threaded Queue Mode => Sequentially adds and releases frames to/from deque and handles overflow of this queue. It utilizes 
	Deques that support thread-safe, memory efficient appends and pops from either side of the deque with approximately the 
	same O(1) performance in either direction.  

	:param source : take the source value. Its default value is 0. Valid Inputs are:

		- Index(integer): Valid index of the video device.

		- YouTube Url(string): Valid Youtube URL as input.

		- Network_Stream_Address(string): Incoming Stream Valid Network address. 

		- GStreamer (string) pipeline

	:param (boolean) y_tube: enables YouTube Mode, i.e If enabled class will interpret the given source string as YouTube URL. Its default value is False.

	:param (int) backend: set the backend of the video stream (if specified). Its default value is 0.

	:param (string) colorspace: set the colorspace of the video stream. Its default value is None.

	:param (dict) **options: provides the ability to tweak properties supported by OpenCV's VideoCapture 
							API properties for any given input video stream directly. All the supported 
							parameters can be passed to CamGear API using this dict as follows:

	:param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.

	:param (integer) time_delay: sets time delay(in seconds) before start reading the frames. 
						/ This delay is essentially required for camera to warm-up. 
						/Its default value is 0.
	"""

	def __init__(self, source = 0, y_tube = False, backend = 0, colorspace = None, logging = False, time_delay = 0, **options):

		#intialize threaded queue mode
		self.threaded_queue_mode = True

		# check if Youtube Mode is ON (True)
		if y_tube:
			try:
				#import pafy and parse youtube stream url
				import pafy
				# validate
				url = youtube_url_validation(source)
				if url:
					source_object = pafy.new(url)
					_source = source_object.getbestvideo("any", ftypestrict=False)
					if _source is None:
						_source = source_object.getbest("any", ftypestrict=False)
					if logging:
						print('URL: {}'.format(url))
						print('Title: {}'.format(source_object.title))
						print('Extension: {}'.format(_source.extension))
					source = _source.url
			except Exception as e:
				if logging:
					print(e)
				raise ValueError('YouTube Mode is enabled and the input YouTube Url is invalid!')

		# youtube mode variable initialization
		self.youtube_mode = y_tube

		#User-Defined Threaded Queue Mode
		if options:
			if "THREADED_QUEUE_MODE" in options:
				if isinstance(options["THREADED_QUEUE_MODE"],bool):
					self.threaded_queue_mode = options["THREADED_QUEUE_MODE"] #assigsn special parameter to global variable
				del options["THREADED_QUEUE_MODE"] #clean
				#reformat option dict

		self.queue = None
		#intialize deque for video files only 
		if self.threaded_queue_mode and isinstance(source,str):
			#import deque
			from collections import deque
			#define deque and assign it to global var
			self.queue = deque(maxlen=96) #max len 96 to check overflow
			#log it
			if logging:
				print('Enabling Threaded Queue Mode for the current video source!') 
		else:
			#otherwise disable it
			self.threaded_queue_mode = False

		# stream variable initialization
		self.stream = None

		if backend and isinstance(backend, int):
			# add backend if specified and initialize the camera stream
			if check_CV_version() == 3:
				# Different OpenCV 3.4.x statement
				self.stream = cv2.VideoCapture(source + backend)
			else:
				# Two parameters are available since OpenCV 4+ (master branch)
				self.stream = cv2.VideoCapture(source, backend)
		else:
			# initialize the camera stream
			self.stream = cv2.VideoCapture(source)


		#initializing colorspace variable
		self.color_space = None

		try: 
			# try to apply attributes to source if specified
			#reformat dict
			options = {k.strip(): v for k,v in options.items()}
			for key, value in options.items():
				self.stream.set(capPropId(key),value)

			# separately handle colorspace value to int conversion
			if not(colorspace is None):
				self.color_space = capPropId(colorspace.strip())

		except Exception as e:
			# Catch if any error occurred
			if logging:
				print(e)

		#initialize and assign framerate variable
		self.framerate = 0
		try:
			_fps = self.stream.get(cv2.CAP_PROP_FPS)
			if _fps>1:
				self.framerate = _fps
		except Exception as e:
			if logging:
				print(e)
			self.framerate = 0

		#frame variable initialization
		(grabbed, self.frame) = self.stream.read()

		if self.threaded_queue_mode:
			#intitialize and append to queue
			self.queue.append(self.frame)

		# applying time delay to warm-up webcam only if specified
		if time_delay:
			time.sleep(time_delay)

		# enable logging if specified
		self.logging = logging

		# thread initialization
		self.thread=None

		# initialize termination flag
		self.terminate = False



	def start(self):
		"""
		start the thread to read frames from the video stream
		"""
		self.thread = Thread(target=self.update, args=())
		self.thread.daemon = True
		self.thread.start()
		return self



	def update(self):
		"""
		Update frames from stream
		"""

		# keep iterating infinitely until the thread is terminated or frames runs out
		while True:
			# if the thread indicator variable is set, stop the thread
			if self.terminate:
				break

			if self.threaded_queue_mode:
				#check queue buffer for overflow
				if len(self.queue) < 96:
					pass
				else:
					#stop iterating if overflowing occurs
					time.sleep(0.000001)
					continue

			# otherwise, read the next frame from the stream
			(grabbed, frame) = self.stream.read()

			#check for valid frames
			if not grabbed:
				#no frames received, then safely exit
				if self.threaded_queue_mode:
					if len(self.queue)>0:
						pass
					else:
						self.terminate = True
				else:
					self.terminate = True

			if not(self.color_space is None):
				# apply colorspace to frames
				color_frame = None
				try:
					if isinstance(self.color_space, int):
						color_frame = cv2.cvtColor(frame, self.color_space)
					else:
						self.color_space = None
						if self.logging:
							print('Colorspace value {} is not a valid Colorspace!'.format(self.color_space))
				except Exception as e:
					# Catch if any error occurred
					self.color_space = None
					if self.logging:
						print(e)
						print('Input Colorspace is not a valid Colorspace!')

				if not(color_frame is None):
					self.frame = color_frame
				else:
					self.frame = frame
			else:
				self.frame = frame

			#append to queue
			if self.threaded_queue_mode:
				self.queue.append(self.frame)

		#release resources
		self.stream.release()



	def read(self):
		"""
		return the frame
		"""
		while self.threaded_queue_mode:
			if len(self.queue)>0:
				return self.queue.popleft()
		return self.frame



	def stop(self):
		"""
		Terminates the Read process
		"""

		#terminate Threaded queue mode seperately
		if self.threaded_queue_mode and not(self.queue is None):
			self.queue.clear()
			self.threaded_queue_mode = False
			self.frame = None

		# indicate that the thread should be terminate
		self.terminate = True
		# wait until stream resources are released (producer thread might be still grabbing frame)
		if self.thread is not None:
			self.thread.join()
			#properly handle thread exit
			if self.youtube_mode:
				# kill thread-lock in youtube mode
				self.thread = None


"""
============================================
vidgear library code is placed under the MIT license
Copyright (c) 2019 Abhishek Thakur

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
===============================================
"""

import platform
import setuptools
from pkg_resources import parse_version
from setuptools import setup

def test_opencv():
    """
    This function is workaround to 
    test if correct OpenCV Library version has already been installed
    on the machine or not. Returns True if previously not installed.
    """
    try:
        # import OpenCV Binaries
        import cv2
        
        if parse_version(cv2.__version__) >= parse_version('3'):
            # check whether OpenCV Binaries are 3.x+
            pass
        else:
            raise ImportError('Incompatible (< 3.0) OpenCV version-{} Installation found on this machine!'.format(parse_version(cv2.__version__)))
    except ImportError:
        return True
    return False

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='vidgear',
    packages=['vidgear','vidgear.gears'],
    version='0.1.5',
    description='Powerful python Video Processing library built with Multi-Threaded Gears(a.k.a APIs) each with a unique set of trailblazing features.',
    license='MIT License',
    author='abhiTronix',
    install_requires = ["pafy", "mss", "youtube-dl", "requests","pyzmq"] 
    + (["opencv-contrib-python"] if test_opencv() else []) 
    + (["picamera"] if ("arm" in platform.uname()[4][:3]) else []),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email='abhi.una12@gmail.com',
    url='https://github.com/abhiTronix/vidgear',
    download_url='https://github.com/abhiTronix/vidgear/tarball/0.1.5',
    keywords=['opencv', 'multithreading', 'FFmpeg', 'picamera', 'mss', 'pyzmq', 'pafy', 'Video Processing', 'Video Stablization', 'Computer Vision'],
    classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7'],
    python_requires='>=2.7',
    scripts=[],
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/abhiTronix/vidgear/issues',
        'Funding': 'https://www.buymeacoffee.com/2twOXFvlA',
        'Source': 'https://github.com/abhiTronix/vidgear',},
)
